/*
 * zr_app.c: ZRadio Main Applicationmodule
 *
 * ZRadio is my very first GTK Application if you have any suggest for my to improve my code
 * please send to my or feel free to include them.
 *
 * Copyright (C) 2014  Matthias Zartmann.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
 * USA
 *
 * Authors:
 *     Matthias Zartmann <matthias.zartmann@gmx.de>
 * Created on: 4 Apr 2014
 */
#include "zr_app.h"
#include "zr_tray.h"
#include "zr_gst.h"
#include "zr_cfg.h"
#include "../config.h"
#include <libnotify/notify.h>

/**
 * Holds the global instances
 */
struct _zr_app
{
	GtkWidget * app_window;   //The hidden Main Mindow
	zr_tray * zr_tray;			 	//The tray
	zr_gst * zr_gst;					//The Gstreamer
	zr_cfg * zr_cfg;          //The configuration
	NotifyNotification *notify;  //The Notify
	gint next_process_steps;  // Bitflags for the enum zr_process_step

	int currentVolume;				//the current Volume
	int argc;									//Application Parameter
	char **argv;

};
typedef struct _zr_app zr_app;

static zr_app G_ZR_APP;  //The only global Variable

/**
 * Shows the notify window
 *
 * @param p_title Title of the Notifywindow
 * @param p_body  Body contens
 * @param time_ms time in ms how long show
 */
static void zr_app_show_notify(gchar * p_title, gchar * p_body, int time_ms)
{

	GError * error = NULL;
	if (G_ZR_APP.notify == NULL)
		G_ZR_APP.notify = notify_notification_new(p_title, p_body, GTK_STOCK_INFO);
	else
		notify_notification_update(G_ZR_APP.notify, p_title, p_body, GTK_STOCK_INFO);

	notify_notification_set_timeout(G_ZR_APP.notify, time_ms);
	if (!notify_notification_show(G_ZR_APP.notify, &error))
	{
		g_error("Failed to send notification. %s", error->message);
	}

}
/**
 * Write the User Config
 */
static void zr_app_writeConfig()
{
	g_debug("Begin writeconfig");
	GKeyFile * inifile = zr_cfg_begin_write();
	gchar * sint = g_strdup_printf("%i", G_ZR_APP.currentVolume);
	zr_cfg_write_value(inifile, ZR_CFG_SEC_CONFIG, ZR_CFG_KEY_VOLUME, sint);
	g_free(sint);
	zr_cfg_write_value(inifile, ZR_CFG_SEC_CONFIG, ZR_CFG_KEY_LASTURL, G_ZR_APP.zr_gst->CurrentPlayingURL);
	zr_cfg_end_write(inifile);
	g_debug("Writeconfig done");
}


/**
 * Determine the last played URL
 * @return Last played URL or NULL
 */
static gchar * zr_app_get_played_last_url(void)
{
	//Use last played
	if (strlen(G_ZR_APP.zr_gst->CurrentPlayingURL) > 0)
	{
		return G_ZR_APP.zr_gst->CurrentPlayingURL;
	}

	//Last Played on exit
	if (G_ZR_APP.zr_cfg->lasturl != NULL && strlen(G_ZR_APP.zr_cfg->lasturl) > 8)
	{
		return G_ZR_APP.zr_cfg->lasturl;
	}
	zr_cfg_station * station = &g_array_index(G_ZR_APP.zr_cfg->stations, zr_cfg_station, 0);
	if (station->key != NULL)
	{

		return station->url;

	}
	return NULL;

}

/**
 * ZRadio internal messages receiver
 * @param gapp The Application
 * @param arg1 the Eventdata
 * @param user_data always null
 */
static void zr_app_messages(GApplication * gapp, zr_app_event * arg1, gpointer user_data)
{
	gchar * tooltip;

	zr_gst_info info;

	g_debug("Callback %s\n", arg1->Message);
	switch (arg1->MessageID)
	{
		case ZR_APP_MSG_BUFFER_EMPTY:
		{

			g_debug("ZR_APP_MSG_BUFFER_EMPTY");
			zr_gst_pause(G_ZR_APP.zr_gst);

			G_ZR_APP.next_process_steps = ZR_PRC_WAIT_INET | ZR_PRC_RECONNECT;

		}
			break;

		case ZR_APP_MSG_CLOSE:
			gtk_widget_destroy(G_ZR_APP.app_window);

			break;

		case ZR_APP_MSG_PLAY:

			if (g_strcmp0(arg1->Message, "LAST") == 0)
			{
				gchar * url = zr_app_get_played_last_url();
				if (url != NULL)
					zr_app_emit_event(ZR_APP_MSG_PLAY, url);
				else
				{
					g_message("No Station to play.");
				}
			} else
			{
				zr_gst_play(G_ZR_APP.zr_gst, arg1->Message);
				zr_gst_set_volume(G_ZR_APP.zr_gst, G_ZR_APP.currentVolume);
				zr_app_writeConfig();
			}
			break;

		case ZR_APP_MSG_PAUSE:
			zr_gst_pause(G_ZR_APP.zr_gst);
			zr_app_show_notify("ZRadio", "Pause", 3000);

			break;

		case ZR_APP_MSG_ICY_CHANGE:

			zr_gst_get_info(G_ZR_APP.zr_gst, &info);
			if (info.playing)
			{
				tooltip = g_strdup_printf("%s\nStation: %s\nSong: %s",PACKAGE_STRING, info.icy_station, info.icy_Song);
			} else
			{
				tooltip = g_strdup_printf("%s\nStopped\nStation: %s\nSong: %s",PACKAGE_STRING, info.icy_station, info.icy_Song);

			}
			zr_tray_set_tooltip(G_ZR_APP.zr_tray, tooltip);

			if (g_strcmp0(arg1->Message, "SONG") == 0)
			{
				zr_app_show_notify("ZRadio", tooltip, 3000);
			}
			g_free(tooltip);
			break;

		case ZR_APP_VOL_CHANGE:
			if (g_strcmp0(arg1->Message, "UP") == 0)
			{
				if (G_ZR_APP.currentVolume <= 90)
				{
					G_ZR_APP.currentVolume += 10;
					zr_gst_set_volume(G_ZR_APP.zr_gst, G_ZR_APP.currentVolume);
					gchar * text = g_strdup_printf("Volume %d %%", G_ZR_APP.currentVolume);
					zr_app_show_notify("ZRadio Volume-Up", text, 3000);
					g_free(text);
				}

			} else if (g_strcmp0(arg1->Message, "DOWN") == 0)
			{

				if (G_ZR_APP.currentVolume > 9)
				{
					G_ZR_APP.currentVolume -= 10;
					zr_gst_set_volume(G_ZR_APP.zr_gst, G_ZR_APP.currentVolume);
					gchar * text = g_strdup_printf("Volume %d %%", G_ZR_APP.currentVolume);
					zr_app_show_notify("ZRadio  Volume-Down", text, 3000);
					g_free(text);

				}
			}

			break;
		case ZR_APP_MSG_RESET:
			G_ZR_APP.next_process_steps = ZR_PRC_WAIT_INET | ZR_PRC_RECONNECT;
			//zr_gst_reset(G_ZR_APP.zr_gst,_zr_app_get_played_last_url());
			break;
		case ZR_APP_MSG_STREAM_ERROR:
			G_ZR_APP.next_process_steps = ZR_PRC_WAIT_INET | ZR_PRC_RECONNECT;
			zr_app_show_notify("ZRadio  Stream-Error", (gchar *) user_data, 3000);

			break;

	}
	g_free(arg1);  //Free the Eventdata
}

/**
 * Creates a new ZRadio internal Eventdata
 * @param pMsgID Messageid
 * @param Message the Message
 * @return allocated Message use G_free
 */
static zr_app_event * zr_app_event_new(enum zr_app_msg pMsgID, gchar * Message)
{
	zr_app_event* result;
	result = g_new(zr_app_event, 1);
	result->MessageID = pMsgID;
	result->Message = Message;
	return result;

}

/**
 * Send a new Event to the internal zr_app_messages
 * @param pMsgID ID Of the Event
 * @param Message the Message data or NULL
 */
void zr_app_emit_event(enum zr_app_msg pMsgID, gchar * Message)
{

	GApplication * gapp = g_application_get_default();
	zr_app_event * event = zr_app_event_new(pMsgID, Message);
	g_signal_emit_by_name(gapp, ZR_APP_EVENT_NAME, event);

}

/**
 * Thats the main Worker of the application called every 1 second
 * @param data ignore
 * @return See GTK
 */

static gboolean zr_app_worker(gpointer * data)
{

	//Must we check the Inet Connection
	if ((G_ZR_APP.next_process_steps & ZR_PRC_WAIT_INET) == ZR_PRC_WAIT_INET)
	{
		GNetworkMonitor * monitor = g_network_monitor_get_default();
		if (!g_network_monitor_get_network_available(monitor))
			return TRUE;
		else
			G_ZR_APP.next_process_steps = G_ZR_APP.next_process_steps & ~ZR_PRC_WAIT_INET;
	}

	//Is this the application Startup

	if ((G_ZR_APP.next_process_steps & ZR_PRC_STARTUP) == ZR_PRC_STARTUP)
	{

		if (G_ZR_APP.zr_cfg->play_on_start)
		{
			zr_gst_play(G_ZR_APP.zr_gst, zr_app_get_played_last_url());
			zr_gst_set_volume(G_ZR_APP.zr_gst, G_ZR_APP.currentVolume);
		}

		G_ZR_APP.next_process_steps = G_ZR_APP.next_process_steps & ~ZR_PRC_STARTUP;
		return TRUE;
	}

	//need a reconnect
	if ((G_ZR_APP.next_process_steps & ZR_PRC_RECONNECT) == ZR_PRC_RECONNECT)
	{
		if (zr_gst_is_playing(G_ZR_APP.zr_gst) == FALSE)
		{
			g_debug("Reconnect");
			//Reconnect to Station Buffer underrun
			zr_gst_reset(G_ZR_APP.zr_gst, zr_app_get_played_last_url());
			G_ZR_APP.next_process_steps = G_ZR_APP.next_process_steps & ~ZR_PRC_RECONNECT;
			return TRUE;
		}
	}

	return TRUE;
}

static void zr_app_activate(GApplication *app, gpointer user_data)
{
	g_debug("activate");
}

/**
 * GTK Callback to startup the Application
 * @param app  See GTK
 * @param user_data always null
 */
static void zr_app_startup(GApplication *app, gpointer user_data)
{

	g_message("startup \n");
	if (g_application_get_is_remote(G_APPLICATION(app)) == TRUE)
	{
		g_message("Another instance of ZRadio is running.");
	}

	gchar * progdir = g_path_get_dirname(G_ZR_APP.argv[0]);
	G_ZR_APP.zr_cfg = zr_cfg_create(progdir);
	g_free(progdir);


	notify_init("Z_RADIO");

	G_ZR_APP.app_window = gtk_application_window_new(GTK_APPLICATION(app));
	gtk_window_set_title(GTK_WINDOW(G_ZR_APP.app_window), "Z_RADIO");

	g_timeout_add(1000, (GSourceFunc) zr_app_worker, NULL);

	//Signal zr_main
	g_signal_new(ZR_APP_EVENT_NAME, G_TYPE_OBJECT, G_SIGNAL_RUN_FIRST, 0, NULL, NULL, g_cclosure_marshal_VOID__POINTER,
	G_TYPE_NONE, 1, G_TYPE_POINTER);
	g_signal_connect(app, ZR_APP_EVENT_NAME, G_CALLBACK(zr_app_messages), NULL);

	G_ZR_APP.zr_tray = zr_tray_create(G_ZR_APP.zr_cfg);
	zr_tray_show(G_ZR_APP.zr_tray);

	G_ZR_APP.zr_gst = zr_gst_create();

	G_ZR_APP.currentVolume = G_ZR_APP.zr_cfg->volume;

	G_ZR_APP.next_process_steps = ZR_PRC_WAIT_INET | ZR_PRC_STARTUP;



}




/**
 * Application Shutdown (Free all resources)
 * @param p_gapp GApplication instance
 */
static void zr_app_shutdown(GApplication * p_gapp)
{
	g_message("Shutdown");

	if (G_ZR_APP.notify != NULL)
	{
		g_object_unref(G_OBJECT(G_ZR_APP.notify));
	}
	notify_uninit();

	zr_tray_free(G_ZR_APP.zr_tray);
	zr_gst_free(G_ZR_APP.zr_gst);
	zr_cfg_free(G_ZR_APP.zr_cfg);

}

/**
 * Main entry for the Apllication
 * @param argc see GNUC
 * @param argv see GNUC
 * @return Application Resultcode
 */
gint zr_app_run(int argc, char *argv[])
{
	G_ZR_APP.argc = argc;
	G_ZR_APP.argv = argv;
	//GError *gerror = NULL;

	GtkApplication * gapp = gtk_application_new("de.ma-za.zr_radio", G_APPLICATION_FLAGS_NONE);

	g_signal_connect(gapp, "startup", G_CALLBACK (zr_app_startup), NULL);
	g_signal_connect(gapp, "activate", G_CALLBACK (zr_app_activate), NULL);
	g_signal_connect(gapp, "shutdown", G_CALLBACK (zr_app_shutdown), NULL);

	// g_application_set_inactivity_timeout (G_APPLICATION (gapp), 10000);
	gint status = g_application_run(G_APPLICATION(gapp), argc, argv);
	g_application_release(G_APPLICATION(gapp));

	return status;

}

